from PyQt5.QtWidgets import QPlainTextEdit
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtCore import Qt, pyqtSignal


class CustomPlainTextEdit(QPlainTextEdit):
    submit = pyqtSignal()

    def __init__(self, submit_callback, parent=None):
        super().__init__(parent)
        self.submit.connect(submit_callback)
        self.setStyleSheet(
            """
            QPlainTextEdit {
                border: none;
                border-radius: 20px;
                padding: 10px 15px;
            }
            """
        )

    def keyPressEvent(self, event: QKeyEvent):
        if (
            event.key() == Qt.Key.Key_Return
            and not event.modifiers() & Qt.KeyboardModifier.ShiftModifier
        ):
            self.submit.emit()
        else:
            super().keyPressEvent(event)



from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtSvg import QSvgRenderer
from PyQt5.QtCore import QByteArray
from PyQt5.QtGui import QPixmap, QPainter

# Updated SVG icons with white fill and stroke
copy_icon_svg = """
<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="white" stroke-width="2" stroke-linecap="round" stroke-linejoin="round">
  <rect x="9" y="9" width="13" height="13" rx="2" ry="2"></rect>
  <path d="M5 15H4a2 2 0 0 1-2-2V4a2 2 0 0 1 2-2h9a2 2 0 0 1 2 2v1"></path>
</svg>
"""

clear_icon_svg = """
<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="white" stroke-width="2" stroke-linecap="round" stroke-linejoin="round">
  <circle cx="12" cy="12" r="10"></circle>
  <line x1="15" y1="9" x2="9" y2="15"></line>
  <line x1="9" y1="9" x2="15" y2="15"></line>
</svg>
"""

microphone_icon_svg = """
<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="#fff" stroke-width="2" stroke-linecap="round" stroke-linejoin="round">
  <path d="M12 1a3 3 0 0 0-3 3v8a3 3 0 0 0 6 0V4a3 3 0 0 0-3-3z"></path>
  <path d="M19 10v2a7 7 0 0 1-14 0v-2"></path>
  <line x1="12" y1="19" x2="12" y2="23"></line>
  <line x1="8" y1="23" x2="16" y2="23"></line>
</svg>
"""


def create_icon_from_svg(svg_string):
    svg_bytes = QByteArray(svg_string.encode("utf-8"))
    renderer = QSvgRenderer(svg_bytes)
    pixmap = QPixmap(24, 24)  # Size of the icon
    pixmap.fill(Qt.GlobalColor.transparent)
    painter = QPainter(pixmap)
    renderer.render(painter)
    painter.end()
    return QIcon(pixmap)



from PyQt5.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QTextBrowser,
    QPushButton,
    QScrollArea,
    QShortcut,
    QSizePolicy,
    QSpacerItem,
)
from PyQt5.QtCore import (
    Qt,
    QSize,
)
from PyQt5.QtGui import (
    QColor,
    QKeySequence,
)

from llama_assistant.custom_plaintext_editor import CustomPlainTextEdit
from llama_assistant.icons import (
    create_icon_from_svg,
    copy_icon_svg,
    clear_icon_svg,
    microphone_icon_svg,
)


class CustomQTextBrowser(QTextBrowser):
    def __init__(self, parent):
        super().__init__(parent)

        # Apply stylesheet specific to generated text content
        self.document().setDefaultStyleSheet(
            """
            p {
                color: #FFFFFF;
                font-size: 16px;
                line-height: 1.3;
            }
            li {
                line-height: 1.3;
            }
            pre {
                color: #FFFFFF;
                background-color: #4A4949;
                border: 1px solid #686763;
                border-radius: 10px;
                font-size: 15px;
                font-family: Consolas, "Courier New", monospace;
                overflow: hidden;
            }
        """
        )


class UIManager:
    def __init__(self, parent):
        self.parent = parent
        self.init_ui()
        self.update_styles()  # Call update_styles after initializing UI

    def init_ui(self):
        self.parent.setWindowTitle("AI Assistant")
        self.parent.setMinimumSize(600, 700)
        self.parent.setWindowFlags(
            Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint
        )
        self.parent.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        central_widget = QWidget(self.parent)
        self.parent.setCentralWidget(central_widget)

        main_layout = QVBoxLayout(central_widget)
        main_layout.setContentsMargins(10, 10, 10, 10)

        top_layout = QVBoxLayout()  # Changed to QVBoxLayout

        input_layout = QHBoxLayout()
        self.input_field = CustomPlainTextEdit(self.parent.on_submit, self.parent)
        self.input_field.setPlaceholderText("Ask me anything...")
        self.input_field.setAcceptDrops(True)
        self.input_field.setFixedHeight(100)
        self.input_field.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.input_field.dragEnterEvent = self.parent.dragEnterEvent
        self.input_field.dropEvent = self.parent.dropEvent
        input_layout.addWidget(self.input_field)

        self.mic_button = QPushButton(self.parent)
        self.mic_button.setIcon(create_icon_from_svg(microphone_icon_svg))
        self.mic_button.setIconSize(QSize(24, 24))
        self.mic_button.setFixedSize(40, 40)
        self.mic_button.clicked.connect(self.parent.toggle_voice_input)
        self.mic_button.setStyleSheet(
            """
            QPushButton {
                background-color: rgba(100, 100, 100, 200);
                border: none;
                border-radius: 20px;
            }
            QPushButton:hover {
                background-color: rgba(100, 100, 100, 230);
            }
        """
        )
        input_layout.addWidget(self.mic_button)

        close_button = QPushButton("×", self.parent)
        close_button.clicked.connect(self.parent.hide)
        close_button.setStyleSheet(
            """
            QPushButton {
                background-color: rgba(255, 0, 0, 150);
                color: white;
                border: none;
                border-radius: 15px;
                font-size: 20px;
                padding: 5px;
                width: 30px;
                height: 30px;
            }
            QPushButton:hover {
                background-color: rgba(255, 0, 0, 200);
            }
        """
        )
        input_layout.addWidget(close_button)

        top_layout.addLayout(input_layout)

        self.image_layout = QHBoxLayout()
        self.image_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        top_layout.addLayout(self.image_layout)

        # Add file layout
        self.file_layout = QHBoxLayout()
        self.file_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        top_layout.addLayout(self.file_layout)

        button_layout = QHBoxLayout()
        button_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.summarize_button = QPushButton("Summarize", self.parent)
        self.rephrase_button = QPushButton("Rephrase", self.parent)
        self.fix_grammar_button = QPushButton("Fix Grammar", self.parent)
        self.brainstorm_button = QPushButton("Brainstorm", self.parent)
        self.write_email_button = QPushButton("Write Email", self.parent)

        for button in [
            self.summarize_button,
            self.rephrase_button,
            self.fix_grammar_button,
            self.brainstorm_button,
            self.write_email_button,
        ]:
            button.clicked.connect(self.parent.on_task_button_clicked)
            button_layout.addWidget(button)

        top_layout.addLayout(button_layout)
        main_layout.addLayout(top_layout)

        result_layout = QHBoxLayout()
        result_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self.copy_button = QPushButton("Copy Result", self.parent)
        self.copy_button.setIcon(create_icon_from_svg(copy_icon_svg))
        self.copy_button.setIconSize(QSize(18, 18))
        self.copy_button.clicked.connect(self.parent.copy_result)
        self.copy_button.hide()

        self.clear_button = QPushButton("Clear", self.parent)
        self.clear_button.setIcon(create_icon_from_svg(clear_icon_svg))
        self.clear_button.setIconSize(QSize(18, 18))
        self.clear_button.clicked.connect(self.parent.clear_chat)
        self.clear_button.hide()

        result_layout.addWidget(self.copy_button)
        result_layout.addWidget(self.clear_button)

        main_layout.addLayout(result_layout)

        self.scroll_area = QScrollArea(self.parent)
        self.scroll_area.setWidgetResizable(True)  # Allow the widget inside to resize
        self.scroll_area.setMinimumHeight(400)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scroll_area.setStyleSheet(
            """
            QScrollArea {
                border: none;
                background-color: transparent;
                border-radius: 20px;
            }
            QScrollBar:vertical {
                border: none;
                background: rgba(255, 255, 255, 200);
                width: 10px;
                margin: 0px 0px 0px 0px;
                border-radius: 5px;
            }
            QScrollBar::handle:vertical {
                background: rgba(255, 255, 255, 230);
                min-height: 20px;
                border-radius: 5px;
            }
            """
        )

        self.chat_box = CustomQTextBrowser(self.scroll_area)
        self.chat_box.setOpenExternalLinks(True)

        self.scroll_area.setWidget(self.chat_box)
        self.scroll_area.hide()  # Hide the scroll area initially

        # Ensure the scroll area can expand fully in the layout
        self.scroll_area.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        main_layout.addWidget(self.scroll_area)

        self.parent.esc_shortcut = QShortcut(QKeySequence("Esc"), self.parent)
        self.parent.esc_shortcut.activated.connect(self.parent.hide)

        # Add an expanding spacer
        spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)
        main_layout.addItem(spacer)

    def update_styles(self):
        opacity = self.parent.settings.get("transparency", 90) / 100
        base_style = f"""
            border: none;
            border-radius: 20px;
            color: white;
            padding: 10px 15px;
            font-size: 16px;
        """
        self.input_field.setStyleSheet(
            f"""
            QPlainTextEdit {{
                background-color: rgba{QColor(self.parent.settings["color"]).getRgb()[:3] + (opacity,)};
                {base_style}
            }}
            """
        )
        self.chat_box.setStyleSheet(
            f"""QTextBrowser {{ {base_style}
                                    background-color: rgba{QColor(self.parent.settings["color"]).lighter(120).getRgb()[:3] + (opacity,)};
                                    border-radius: 10px;
                                    }}"""
        )
        button_style = f"""
            QPushButton {{
                {base_style}
                padding: 2.5px 5px;
                border-radius: 5px;
                background-color: rgba{QColor(self.parent.settings["color"]).getRgb()[:3] + (opacity,)};
            }}
            QPushButton:hover {{
                background-color: rgba{QColor(self.parent.settings["color"]).lighter(120).getRgb()[:3] + (opacity,)};
            }}
        """
        for button in [
            self.rephrase_button,
            self.fix_grammar_button,
            self.brainstorm_button,
            self.write_email_button,
            self.summarize_button,
        ]:
            button.setStyleSheet(button_style)

        button_style = f"""
            QPushButton {{
                {base_style}
                padding: 2.5px 5px;
                border-radius: 5px;
                background-color: rgba{QColor(self.parent.settings["color"]).lighter(120).getRgb()[:3] + (opacity,)};
            }}
            QPushButton:hover {{
                background-color: rgba{QColor(self.parent.settings["color"]).lighter(150).getRgb()[:3] + (opacity,)};
            }}
        """
        for button in [self.copy_button, self.clear_button]:
            button.setStyleSheet(button_style)



import json
from pathlib import Path
from PyQt5.QtWidgets import (
    QDialog,
    QFormLayout,
    QPushButton,
    QSlider,
    QComboBox,
    QColorDialog,
    QVBoxLayout,
    QHBoxLayout,
    QCheckBox,
    QGroupBox,
    QLineEdit,
    QMessageBox,
    QListWidget,
    QLabel,
)
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QColor
from pynput import keyboard

from llama_assistant.shortcut_recorder import ShortcutRecorder
from llama_assistant import config


class SettingsDialog(QDialog):
    settingsSaved = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Settings")
        self.main_layout = QVBoxLayout(self)

        # General Settings Group
        self.create_general_settings_group()

        # Appearance Settings Group
        self.create_appearance_settings_group()

        # Model Settings Group
        self.create_model_settings_group()

        # Voice Activation Settings Group
        self.create_voice_activation_settings_group()

        self.create_rag_settings_group()

        # Create a horizontal layout for the save button
        button_layout = QHBoxLayout()
        self.save_button = QPushButton("Save")
        self.save_button.clicked.connect(self.accept)
        button_layout.addStretch()
        button_layout.addWidget(self.save_button)

        # Add the button layout to the main layout
        self.main_layout.addLayout(button_layout)

        self.load_settings()

    def create_general_settings_group(self):
        group_box = QGroupBox("General Settings")
        layout = QVBoxLayout()

        shortcut_layout = QHBoxLayout()
        shortcut_label = QLabel("Shortcut:")
        self.shortcut_recorder = ShortcutRecorder()
        shortcut_layout.addWidget(shortcut_label)
        shortcut_layout.addWidget(self.shortcut_recorder)
        shortcut_layout.addStretch()
        layout.addLayout(shortcut_layout)

        self.reset_shortcut_button = QPushButton("Reset Shortcut")
        self.reset_shortcut_button.clicked.connect(self.reset_shortcut)
        layout.addWidget(self.reset_shortcut_button)

        group_box.setLayout(layout)
        self.main_layout.addWidget(group_box)

    def create_appearance_settings_group(self):
        group_box = QGroupBox("Appearance Settings")
        layout = QVBoxLayout()

        color_layout = QHBoxLayout()
        color_label = QLabel("Background Color:")
        self.color_button = QPushButton("Choose Color")
        self.color_button.clicked.connect(self.choose_color)
        color_layout.addWidget(color_label)
        color_layout.addWidget(self.color_button)
        color_layout.addStretch()
        layout.addLayout(color_layout)

        transparency_layout = QHBoxLayout()
        transparency_label = QLabel("Transparency:")
        self.transparency_slider = QSlider(Qt.Orientation.Horizontal)
        self.transparency_slider.setRange(10, 100)
        self.transparency_slider.setValue(90)
        transparency_layout.addWidget(transparency_label)
        transparency_layout.addWidget(self.transparency_slider)
        layout.addLayout(transparency_layout)

        group_box.setLayout(layout)
        self.main_layout.addWidget(group_box)

    def create_model_settings_group(self):
        group_box = QGroupBox("Model Settings")
        layout = QVBoxLayout()

        text_model_layout = QHBoxLayout()
        text_model_label = QLabel("Text-only Model:")
        self.text_model_combo = QComboBox()
        self.text_model_combo.addItems(self.get_model_names_by_type("text"))
        text_model_layout.addWidget(text_model_label)
        text_model_layout.addWidget(self.text_model_combo)
        text_model_layout.addStretch()
        layout.addLayout(text_model_layout)

        multimodal_model_layout = QHBoxLayout()
        multimodal_model_label = QLabel("Multimodal Model:")
        self.multimodal_model_combo = QComboBox()
        self.multimodal_model_combo.addItems(self.get_model_names_by_type("image"))
        multimodal_model_layout.addWidget(multimodal_model_label)
        multimodal_model_layout.addWidget(self.multimodal_model_combo)
        multimodal_model_layout.addStretch()
        layout.addLayout(multimodal_model_layout)

        context_len_layout = QHBoxLayout()
        context_len_label = QLabel("Context Length:")
        self.context_len_input = QLineEdit()
        context_len_layout.addWidget(context_len_label)
        context_len_layout.addWidget(self.context_len_input)
        layout.addLayout(context_len_layout)

        temperature_layout = QHBoxLayout()
        temperature_label = QLabel("Temperature:")
        self.temperature_input = QLineEdit()
        temperature_layout.addWidget(temperature_label)
        temperature_layout.addWidget(self.temperature_input)
        layout.addLayout(temperature_layout)

        top_p_layout = QHBoxLayout()
        top_p_label = QLabel("Top p:")
        self.top_p_input = QLineEdit()
        top_p_layout.addWidget(top_p_label)
        top_p_layout.addWidget(self.top_p_input)
        layout.addLayout(top_p_layout)

        top_k_layout = QHBoxLayout()
        top_k_label = QLabel("Top k:")
        self.top_k_input = QLineEdit()
        top_k_layout.addWidget(top_k_label)
        top_k_layout.addWidget(self.top_k_input)
        layout.addLayout(top_k_layout)

        self.manage_custom_models_button = QPushButton("Manage Custom Models")
        self.manage_custom_models_button.clicked.connect(self.open_custom_models_dialog)
        layout.addWidget(self.manage_custom_models_button)

        group_box.setLayout(layout)
        self.main_layout.addWidget(group_box)

    def create_voice_activation_settings_group(self):
        group_box = QGroupBox("Voice Activation Settings")
        layout = QVBoxLayout()

        self.hey_llama_chat_checkbox = QCheckBox('Say "Hey Llama" to open chat form')
        self.hey_llama_chat_checkbox.stateChanged.connect(self.update_hey_llama_mic_state)
        layout.addWidget(self.hey_llama_chat_checkbox)

        self.hey_llama_mic_checkbox = QCheckBox('Say "Hey Llama" to activate microphone')
        layout.addWidget(self.hey_llama_mic_checkbox)

        group_box.setLayout(layout)
        self.main_layout.addWidget(group_box)

    def create_rag_settings_group(self):
        group_box = QGroupBox("RAG Settings")
        layout = QVBoxLayout()

        embed_model_layout = QHBoxLayout()
        embed_model_label = QLabel("Embed Model Name:")
        self.embed_model_input = QLineEdit()
        embed_model_layout.addWidget(embed_model_label)
        embed_model_layout.addWidget(self.embed_model_input)
        layout.addLayout(embed_model_layout)

        chunk_size_layout = QHBoxLayout()
        chunk_size_label = QLabel("Chunk Size:")
        self.chunk_size_input = QLineEdit()
        chunk_size_layout.addWidget(chunk_size_label)
        chunk_size_layout.addWidget(self.chunk_size_input)
        layout.addLayout(chunk_size_layout)

        chunk_overlap_layout = QHBoxLayout()
        chunk_overlap_label = QLabel("Chunk Overlap:")
        self.chunk_overlap_input = QLineEdit()
        chunk_overlap_layout.addWidget(chunk_overlap_label)
        chunk_overlap_layout.addWidget(self.chunk_overlap_input)
        layout.addLayout(chunk_overlap_layout)

        max_retrieval_top_k_layout = QHBoxLayout()
        max_retrieval_top_k_label = QLabel("Max Retrieval Top k:")
        self.max_retrieval_top_k_input = QLineEdit()
        max_retrieval_top_k_layout.addWidget(max_retrieval_top_k_label)
        max_retrieval_top_k_layout.addWidget(self.max_retrieval_top_k_input)
        layout.addLayout(max_retrieval_top_k_layout)

        similarity_threshold_layout = QHBoxLayout()
        similarity_threshold_label = QLabel("Similarity Threshold:")
        self.similarity_threshold_input = QLineEdit()
        similarity_threshold_layout.addWidget(similarity_threshold_label)
        similarity_threshold_layout.addWidget(self.similarity_threshold_input)
        layout.addLayout(similarity_threshold_layout)

        group_box.setLayout(layout)
        self.main_layout.addWidget(group_box)

    def accept(self):
        self.save_settings()
        self.settingsSaved.emit()
        super().accept()

    def get_model_names_by_type(self, model_type):
        return [model["model_id"] for model in config.models if model["model_type"] == model_type]

    def choose_color(self):
        color = QColorDialog.getColor()
        if color.isValid():
            self.color = color

    def reset_shortcut(self):
        self.shortcut_recorder.setText(config.DEFAULT_LAUNCH_SHORTCUT)

    def update_hey_llama_mic_state(self, state):
        self.hey_llama_mic_checkbox.setEnabled(state == Qt.Checked)

    def load_settings(self):
        if config.settings_file.exists():
            with open(config.settings_file, "r") as f:
                settings = json.load(f)
            try:
                keyboard.HotKey(keyboard.HotKey.parse(settings["shortcut"]), lambda: None)
            except ValueError:
                settings["shortcut"] = config.DEFAULT_LAUNCH_SHORTCUT
                self.save_settings(settings)
            self.shortcut_recorder.setText(settings.get("shortcut", config.DEFAULT_LAUNCH_SHORTCUT))
            self.color = QColor(settings.get("color", "#1E1E1E"))
            self.transparency_slider.setValue(int(settings.get("transparency", 90)))

            text_model = settings.get("text_model")
            if text_model in self.get_model_names_by_type("text"):
                self.text_model_combo.setCurrentText(text_model)

            multimodal_model = settings.get("multimodal_model")
            if multimodal_model in self.get_model_names_by_type("image"):
                self.multimodal_model_combo.setCurrentText(multimodal_model)

            self.hey_llama_chat_checkbox.setChecked(settings.get("hey_llama_chat", False))
            self.hey_llama_mic_checkbox.setChecked(settings.get("hey_llama_mic", False))
            self.update_hey_llama_mic_state(settings.get("hey_llama_chat", False))

            # Load new settings
            if "generation" not in settings:
                settings["generation"] = {}
            if "rag" not in settings:
                settings["rag"] = {}
            self.embed_model_input.setText(
                settings["rag"].get("embed_model_name", config.embed_model_name)
            )
            self.chunk_size_input.setText(str(settings["rag"].get("chunk_size", config.chunk_size)))
            self.chunk_overlap_input.setText(
                str(settings["rag"].get("chunk_overlap", config.chunk_overlap))
            )
            self.max_retrieval_top_k_input.setText(
                str(settings["rag"].get("max_retrieval_top_k", config.max_retrieval_top_k))
            )
            self.similarity_threshold_input.setText(
                str(settings["rag"].get("similarity_threshold", config.similarity_threshold))
            )
            self.context_len_input.setText(
                str(settings["generation"].get("context_len", config.context_len))
            )
            self.temperature_input.setText(
                str(settings["generation"].get("temperature", config.temperature))
            )
            self.top_p_input.setText(str(settings["generation"].get("top_p", config.top_p)))
            self.top_k_input.setText(str(settings["generation"].get("top_k", config.top_k)))
        else:
            self.color = QColor("#1E1E1E")
            self.shortcut_recorder.setText("<cmd>+<shift>+<space>")

    def get_settings(self):
        return {
            "shortcut": self.shortcut_recorder.text(),
            "color": self.color.name(),
            "transparency": self.transparency_slider.value(),
            "text_model": self.text_model_combo.currentText(),
            "multimodal_model": self.multimodal_model_combo.currentText(),
            "hey_llama_chat": self.hey_llama_chat_checkbox.isChecked(),
            "hey_llama_mic": self.hey_llama_mic_checkbox.isChecked(),
            "generation": {
                "context_len": int(self.context_len_input.text()),
                "temperature": float(self.temperature_input.text()),
                "top_p": float(self.top_p_input.text()),
                "top_k": int(self.top_k_input.text()),
            },
            "rag": {
                "embed_model_name": self.embed_model_input.text(),
                "chunk_size": int(self.chunk_size_input.text()),
                "chunk_overlap": int(self.chunk_overlap_input.text()),
                "max_retrieval_top_k": int(self.max_retrieval_top_k_input.text()),
                "similarity_threshold": float(self.similarity_threshold_input.text()),
            },
        }

    def save_settings(self, settings=None):
        if settings is None:
            settings = self.get_settings()

        with open(config.settings_file, "w") as f:
            json.dump(settings, f)

    def open_custom_models_dialog(self):
        dialog = CustomModelsDialog(self)
        if dialog.exec():
            # Refresh the model combos after managing custom models
            self.refresh_model_combos()
        self.refresh_model_combos()  # Run refresh_model_combos after closing the custom models editor

    def refresh_model_combos(self):
        current_text_model = self.text_model_combo.currentText()
        current_multimodal_model = self.multimodal_model_combo.currentText()

        self.text_model_combo.clear()
        self.text_model_combo.addItems(self.get_model_names_by_type("text"))
        self.multimodal_model_combo.clear()
        self.multimodal_model_combo.addItems(self.get_model_names_by_type("image"))

        # Restore previously selected models if they still exist
        if current_text_model in self.get_model_names_by_type("text"):
            self.text_model_combo.setCurrentText(current_text_model)
        if current_multimodal_model in self.get_model_names_by_type("image"):
            self.multimodal_model_combo.setCurrentText(current_multimodal_model)


class CustomModelsDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Manage Custom Models")
        self.layout = QVBoxLayout(self)

        self.model_list = QListWidget()
        self.model_list.itemSelectionChanged.connect(self.load_selected_model)
        self.layout.addWidget(self.model_list)

        form_layout = QFormLayout()
        self.model_name_input = QLineEdit()
        self.model_id_input = QLineEdit()
        self.model_type_input = QComboBox()
        self.model_type_input.addItems(["text", "image"])
        self.repo_id_input = QLineEdit()
        self.filename_input = QLineEdit()

        form_layout.addRow("Model Name:", self.model_name_input)
        form_layout.addRow("Model ID:", self.model_id_input)
        form_layout.addRow("Model Type:", self.model_type_input)
        form_layout.addRow("Repo ID:", self.repo_id_input)
        form_layout.addRow("Filename:", self.filename_input)

        self.layout.addLayout(form_layout)

        button_layout = QHBoxLayout()
        self.add_button = QPushButton("Add Model")
        self.add_button.clicked.connect(self.add_model)
        self.update_button = QPushButton("Update Model")
        self.update_button.clicked.connect(self.update_model)
        self.remove_button = QPushButton("Remove Model")
        self.remove_button.clicked.connect(self.remove_model)
        button_layout.addWidget(self.add_button)
        button_layout.addWidget(self.update_button)
        button_layout.addWidget(self.remove_button)

        self.layout.addLayout(button_layout)

        self.refresh_model_list()

    def refresh_model_list(self):
        self.model_list.clear()
        for model in config.custom_models:
            self.model_list.addItem(f"{model['model_name']} ({model['model_type']})")

    def load_selected_model(self):
        selected_items = self.model_list.selectedItems()
        if selected_items:
            selected_index = self.model_list.row(selected_items[0])
            model = config.custom_models[selected_index]
            self.model_name_input.setText(model["model_name"])
            self.model_id_input.setText(model["model_id"])
            self.model_type_input.setCurrentText(model["model_type"])
            self.repo_id_input.setText(model["repo_id"])
            self.filename_input.setText(model["filename"])

    def add_model(self):
        model_name = self.model_name_input.text()
        model_id = self.model_id_input.text()
        model_type = self.model_type_input.currentText()
        repo_id = self.repo_id_input.text()
        filename = self.filename_input.text()

        if not all([model_name, model_id, model_type, repo_id, filename]):
            QMessageBox.warning(self, "Missing Information", "Please fill in all fields.")
            return

        new_model = {
            "model_name": model_name,
            "model_id": model_id,
            "model_type": model_type,
            "model_path": None,
            "repo_id": repo_id,
            "filename": filename,
        }

        config.custom_models.append(new_model)
        config.save_custom_models()
        self.refresh_model_list()
        self.clear_inputs()
        QMessageBox.information(
            self, "Model Added", f"Model '{model_name}' has been added successfully."
        )

    def update_model(self):
        selected_items = self.model_list.selectedItems()
        if not selected_items:
            QMessageBox.warning(self, "No Selection", "Please select a model to update.")
            return

        selected_index = self.model_list.row(selected_items[0])
        model_name = self.model_name_input.text()
        model_id = self.model_id_input.text()
        model_type = self.model_type_input.currentText()
        repo_id = self.repo_id_input.text()
        filename = self.filename_input.text()

        if not all([model_name, model_id, model_type, repo_id, filename]):
            QMessageBox.warning(self, "Missing Information", "Please fill in all fields.")
            return

        updated_model = {
            "model_name": model_name,
            "model_id": model_id,
            "model_type": model_type,
            "model_path": None,
            "repo_id": repo_id,
            "filename": filename,
        }

        config.custom_models[selected_index] = updated_model
        config.models = config.DEFAULT_MODELS + config.custom_models
        config.save_custom_models()
        self.refresh_model_list()
        self.clear_inputs()
        QMessageBox.information(
            self, "Model Updated", f"Model '{model_name}' has been updated successfully."
        )

    def remove_model(self):
        selected_items = self.model_list.selectedItems()
        if not selected_items:
            QMessageBox.warning(self, "No Selection", "Please select a model to remove.")
            return

        selected_index = self.model_list.row(selected_items[0])
        model_name = config.custom_models[selected_index]["model_name"]
        del config.custom_models[selected_index]
        config.models = config.DEFAULT_MODELS + config.custom_models
        config.save_custom_models()
        self.refresh_model_list()
        self.clear_inputs()
        QMessageBox.information(
            self, "Model Removed", f"Model '{model_name}' has been removed successfully."
        )

    def clear_inputs(self):
        self.model_name_input.clear()
        self.model_id_input.clear()
        self.model_type_input.setCurrentIndex(0)
        self.repo_id_input.clear()
        self.filename_input.clear()



import sys

from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QKeyEvent, QKeySequence


def is_macos():
    return sys.platform == "darwin"


class ShortcutRecorder(QLineEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setReadOnly(True)
        self.setPlaceholderText("Press a key combination")
        self.recorded_shortcut = None

        # Set the style sheet for rounded corners
        self.setStyleSheet(
            """
            QLineEdit {
                border: 1px solid #a0a0a0;
                border-radius: 7.5px;
                padding: 2px 5px;
                background-color: #fefefe;
                color: #333333;
                font-size: 14px;
            }
            QLineEdit:focus {
                border-color: #3498db;
            }
        """
        )

    def keyPressEvent(self, event: QKeyEvent):
        modifiers = []
        if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            if is_macos():
                modifiers.append("<cmd>")
            else:
                modifiers.append("<ctrl>")
        if event.modifiers() & Qt.KeyboardModifier.AltModifier:
            modifiers.append("<alt>")
        if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
            modifiers.append("<shift>")
        if event.modifiers() & Qt.KeyboardModifier.MetaModifier:
            if is_macos():
                modifiers.append("<ctrl>")
            else:
                modifiers.append("<cmd>")

        key = event.key()
        if key not in (
            Qt.Key.Key_Control,
            Qt.Key.Key_Shift,
            Qt.Key.Key_Alt,
            Qt.Key.Key_Meta,
        ):
            qt_key = QKeySequence(key).toString()
            pynput_key = self.qt_to_pynput_key(qt_key)
            key_string = "+".join(modifiers + [pynput_key])
            self.setText(key_string)
            self.recorded_shortcut = key_string

        event.accept()

    def qt_to_pynput_key(self, qt_key):
        # Map Qt key names to pynput key names
        key_map = {
            "PgUp": "<page_up>",
            "PgDown": "<page_down>",
            "Left": "<left>",
            "Right": "<right>",
            "Up": "<up>",
            "Down": "<down>",
            "Enter": "<enter>",
            "Return": "<enter>",
            "Ins": "<insert>",
            "Del": "<delete>",
            "Home": "<home>",
            "End": "<end>",
            "Space": "<space>",
            "Tab": "<tab>",
            "Esc": "<esc>",
            "Backspace": "<backspace>",
            # Function keys
            "F1": "<f1>",
            "F2": "<f2>",
            "F3": "<f3>",
            "F4": "<f4>",
            "F5": "<f5>",
            "F6": "<f6>",
            "F7": "<f7>",
            "F8": "<f8>",
            "F9": "<f9>",
            "F10": "<f10>",
            "F11": "<f11>",
            "F12": "<f12>",
        }
        return key_map.get(qt_key, qt_key.lower())

    def get_pynput_hotkey(self):
        return self.recorded_shortcut



from PyQt5.QtCore import QObject, pyqtSignal
from pynput import keyboard

from llama_assistant.config import DEFAULT_LAUNCH_SHORTCUT


class GlobalHotkey(QObject):
    activated = pyqtSignal()

    def __init__(self, hotkey):
        super().__init__()
        try:
            self.hotkey = keyboard.HotKey(keyboard.HotKey.parse(hotkey), self.on_activate)
        except ValueError:
            self.hotkey = keyboard.HotKey(
                keyboard.HotKey.parse(DEFAULT_LAUNCH_SHORTCUT), self.on_activate
            )
        self.listener = keyboard.Listener(
            on_press=self.for_canonical(self.hotkey.press),
            on_release=self.for_canonical(self.hotkey.release),
        )
        self.listener.start()

    def on_activate(self):
        self.activated.emit()

    def for_canonical(self, f):
        return lambda k: f(self.listener.canonical(k))

    def stop(self):
        if self.listener:
            self.listener.stop()



import pyaudio
import numpy as np
from openwakeword.model import Model
from openwakeword.utils import download_models
from PyQt5.QtCore import QThread, pyqtSignal
import time

from llama_assistant.utils import get_resource_path


class WakeWordDetector(QThread):
    wakeword_detected = pyqtSignal(str)

    def __init__(self, chunk_size=1280, inference_framework="onnx"):
        super().__init__()
        self.chunk_size = chunk_size
        self.inference_framework = inference_framework
        self.rate = 16000
        self.format = pyaudio.paInt16
        self.channels = 1
        self.running = False

        self.audio = pyaudio.PyAudio()
        self.mic_stream = self.audio.open(
            format=self.format,
            channels=self.channels,
            rate=self.rate,
            input=True,
            frames_per_buffer=self.chunk_size,
        )

        self.owwModel = None
        self.n_models = 0

    def load_model(self):
        path = get_resource_path("llama_assistant/resources/wk_hey_llama.onnx")
        self.owwModel = Model(
            wakeword_models=[path],
            inference_framework=self.inference_framework,
        )
        self.n_models = len(self.owwModel.models.keys())

    def unload_model(self):
        self.owwModel = None
        self.n_models = 0

    def run(self):
        self.running = True
        download_models()
        self.load_model()
        while self.running:
            try:
                audio = np.frombuffer(self.mic_stream.read(self.chunk_size), dtype=np.int16)
                prediction = self.owwModel.predict(audio)
                self.process_prediction(prediction)
                time.sleep(0.01)  # Small delay to prevent CPU overuse
            except Exception as e:
                print(f"Error: {e}")
                self.stop()
        self.unload_model()

    def process_prediction(self, prediction):
        for mdl, scores in self.owwModel.prediction_buffer.items():
            if scores[-1] > 0.5:
                self.wakeword_detected.emit(mdl)
                print(f"Wakeword detected: {mdl}")

    def stop(self):
        self.running = False
        self.wait()

    def print_results(self):
        if not self.owwModel:
            print("Model not loaded")
            return

        n_spaces = 16
        output_string_header = """
            Model Name         | Score | Wakeword Status
            --------------------------------------
            """

        for mdl in self.owwModel.prediction_buffer.keys():
            scores = list(self.owwModel.prediction_buffer[mdl])
            curr_score = format(scores[-1], ".20f").replace("-", "")

            output_string_header += f"""{mdl}{" "*(n_spaces - len(mdl))}   | {curr_score[0:5]} | {"--"+" "*20 if scores[-1] <= 0.5 else "Wakeword Detected!"}
            """

        print("\033[F" * (4 * self.n_models + 1))
        print(output_string_header, "                             ", end="\r")



from PyQt5.QtWidgets import (
    QApplication,
    QSystemTrayIcon,
    QMenu,
)
from PyQt5.QtGui import (
    QIcon,
)

from llama_assistant.utils import load_image


class TrayManager:
    def __init__(self, parent):
        self.parent = parent
        self.init_tray()

    def init_tray(self):
        self.tray_icon = QSystemTrayIcon(self.parent)
        self.tray_icon.setIcon(self.load_tray_icon())

        tray_menu = QMenu()
        show_action = tray_menu.addAction("Show")
        show_action.triggered.connect(self.parent.show)
        settings_action = tray_menu.addAction("Settings")
        settings_action.triggered.connect(self.parent.open_settings)
        quit_action = tray_menu.addAction("Quit")
        quit_action.triggered.connect(QApplication.quit)

        self.tray_icon.setContextMenu(tray_menu)
        self.tray_icon.activated.connect(self.parent.tray_icon_activated)
        self.tray_icon.show()

    def load_tray_icon(self):
        icon_path = "llama_assistant/resources/logo.png"
        pixmap = load_image(icon_path, size=(48, 48))
        return QIcon(pixmap)



from typing import Set, Optional, Dict
from PyQt5.QtCore import (
    QThread,
    pyqtSignal,
)
from llama_assistant.model_handler import handler as model_handler


class ProcessingThread(QThread):
    preloader_signal = pyqtSignal(str)
    update_signal = pyqtSignal(str)
    finished_signal = pyqtSignal()

    def __init__(
        self,
        model: str,
        generation_setting: Dict,
        rag_setting: Dict,
        prompt: str,
        lookup_files: Optional[Set[str]] = None,
        image: str = None,
    ):
        super().__init__()
        self.model = model
        self.generation_setting = generation_setting
        self.rag_setting = rag_setting
        self.prompt = prompt
        self.image = image
        self.lookup_files = lookup_files
        self.preloadong = False

    def run(self):
        output = model_handler.chat_completion(
            self.model,
            self.generation_setting,
            self.rag_setting,
            self.prompt,
            image=self.image,
            lookup_files=self.lookup_files,
            stream=True,
            processing_thread=self,
        )
        full_response_str = ""
        for chunk in output:
            delta = chunk["choices"][0]["delta"]
            if "role" in delta:
                print(delta["role"], end=": ")
            elif "content" in delta:
                print(delta["content"], end="")
                full_response_str += delta["content"]
                self.update_signal.emit(delta["content"])
        model_handler.update_chat_history(full_response_str, "assistant")
        self.finished_signal.emit()

    def clear_chat_history(self):
        model_handler.clear_chat_history()
        self.finished_signal.emit()

    def emit_preloading_message(self, message: str):
        self.preloader_signal.emit(message)

    def set_preloading(self, preloading: bool, message: str):
        self.preloading = preloading
        self.emit_preloading_message(message)

    def is_preloading(self):
        return self.preloading



import asyncio
from typing import List, Dict, Set, Optional, TYPE_CHECKING
import time
from threading import Timer
from llama_cpp import Llama
from llama_cpp.llama_chat_format import (
    MoondreamChatHandler,
    MiniCPMv26ChatHandler,
    Llava15ChatHandler,
    Llava16ChatHandler,
)

from llama_assistant import config
from llama_assistant.agent import RAGAgent

if TYPE_CHECKING:
    from llama_assistant.processing_thread import ProcessingThread


class Model:
    def __init__(
        self,
        model_type: str,
        model_id: str,
        model_name: str,
        model_path: Optional[str] = None,
        repo_id: Optional[str] = None,
        filename: Optional[str] = None,
    ):
        self.model_type = model_type
        self.model_id = model_id
        self.model_name = model_name
        self.model_path = model_path
        self.repo_id = repo_id
        self.filename = filename

    def is_online(self) -> bool:
        return self.repo_id is not None and self.filename is not None


class ModelHandler:
    def __init__(self):
        self.supported_models: List[Model] = []
        self.loaded_agent: Optional[Dict] = None
        self.current_model_id: Optional[str] = None
        self.unload_timer: Optional[Timer] = None

    def refresh_supported_models(self):
        self.supported_models = [Model(**model_data) for model_data in config.models]

    def list_supported_models(self) -> List[Model]:
        return self.supported_models

    def add_supported_model(self, model: Model):
        self.supported_models.append(model)

    def remove_supported_model(self, model_id: str):
        self.supported_models = [m for m in self.supported_models if m.model_id != model_id]
        if self.current_model_id == model_id:
            self.unload_agent()

    def load_agent(
        self,
        model_id: str,
        generation_setting,
        rag_setting: Dict,
        processing_thread: "ProcessingThread",
    ) -> Optional[Dict]:
        self.refresh_supported_models()
        if self.current_model_id == model_id and self.loaded_agent:
            if generation_setting["context_len"] == self.loaded_agent["model"].context_params.n_ctx:
                self.loaded_agent["agent"].update_rag_setting(rag_setting)
                self.loaded_agent["agent"].update_generation_setting(generation_setting)
                return self.loaded_agent

        processing_thread.set_preloading(True, "Loading model ....")

        # if no model is loaded or different model is loaded, or context_len is different, reinitialize the agent
        self.unload_agent()  # Unload the current model if any

        model = next((m for m in self.supported_models if m.model_id == model_id), None)
        if not model:
            print(f"Model with ID {model_id} not found.")
            return None

        if model.is_online():
            if model.model_type == "text":
                print("load online model")
                loaded_model = Llama.from_pretrained(
                    repo_id=model.repo_id,
                    filename=model.filename,
                    n_ctx=generation_setting["context_len"],
                )
            elif model.model_type == "image":
                if "moondream2" in model.model_id:
                    chat_handler = MoondreamChatHandler.from_pretrained(
                        repo_id="vikhyatk/moondream2",
                        filename="*mmproj*",
                    )
                    loaded_model = Llama.from_pretrained(
                        repo_id=model.repo_id,
                        filename=model.filename,
                        chat_handler=chat_handler,
                        n_ctx=generation_setting["context_len"],
                    )
                elif "MiniCPM" in model.model_id:
                    chat_handler = MiniCPMv26ChatHandler.from_pretrained(
                        repo_id=model.repo_id,
                        filename="*mmproj*",
                    )
                    loaded_model = Llama.from_pretrained(
                        repo_id=model.repo_id,
                        filename=model.filename,
                        chat_handler=chat_handler,
                        n_ctx=generation_setting["context_len"],
                    )
                elif "llava-v1.5" in model.model_id:
                    chat_handler = Llava15ChatHandler.from_pretrained(
                        repo_id=model.repo_id,
                        filename="*mmproj*",
                    )
                    loaded_model = Llama.from_pretrained(
                        repo_id=model.repo_id,
                        filename=model.filename,
                        chat_handler=chat_handler,
                        n_ctx=generation_setting["context_len"],
                    )
                elif "llava-v1.6" in model.model_id:
                    chat_handler = Llava16ChatHandler.from_pretrained(
                        repo_id=model.repo_id,
                        filename="*mmproj*",
                    )
                    loaded_model = Llama.from_pretrained(
                        repo_id=model.repo_id,
                        filename=model.filename,
                        chat_handler=chat_handler,
                        n_ctx=generation_setting["context_len"],
                    )
            else:
                print(f"Unsupported model type: {model.model_type}")
                return None
        else:
            # Load model from local path
            print("load local model")
            loaded_model = Llama(model_path=model.model_path)

        print("Initializing agent ...")

        agent = RAGAgent(
            generation_setting,
            rag_setting,
            llm=loaded_model,
        )

        self.loaded_agent = {
            "model": loaded_model,
            "agent": agent,
            "generation_setting": generation_setting,
            "rag_setting": rag_setting,
            "last_used": time.time(),
        }
        self.current_model_id = model_id
        self._schedule_unload()

        return self.loaded_agent

    def unload_agent(self):
        if self.loaded_agent:
            print(f"Unloading model: {self.current_model_id}")
            self.loaded_agent = None
            self.current_model_id = None
        if self.unload_timer:
            self.unload_timer.cancel()
            self.unload_timer = None

    async def run_agent(
        self, agent: RAGAgent, message: str, lookup_files: Set, image: str, stream: bool
    ):
        response = await agent.run(
            query_str=message, lookup_files=lookup_files, image=image, streaming=stream
        )
        return response

    def chat_completion(
        self,
        model_id: str,
        generation_setting: Dict,
        rag_setting: Dict,
        message: str,
        image: Optional[str] = None,
        lookup_files: Optional[Set[str]] = None,
        stream: bool = False,
        processing_thread: "ProcessingThread" = None,
    ) -> str:
        agent_data = self.load_agent(model_id, generation_setting, rag_setting, processing_thread)
        agent = agent_data.get("agent")
        if not agent_data:
            return "Failed to load model"

        agent_data["last_used"] = time.time()
        self._schedule_unload()

        processing_thread.set_preloading(True, "Thinking ....")
        try:
            loop = asyncio.get_running_loop()
            response = loop.run_until_complete(
                self.run_agent(agent, message, lookup_files, image, stream)
            )
        except RuntimeError:  # no running event loop
            response = asyncio.run(self.run_agent(agent, message, lookup_files, image, stream))

        processing_thread.set_preloading(False, "Thinking done.")

        return response

    def update_chat_history(self, message: str, role: str):
        agent = self.loaded_agent.get("agent")
        if agent:
            agent.chat_history.add_message({"role": role, "content": message})

    def clear_chat_history(self):
        agent = self.loaded_agent.get("agent")
        if agent:
            agent.chat_history.clear()

    def _schedule_unload(self):
        if self.unload_timer:
            self.unload_timer.cancel()

        self.unload_timer = Timer(3600, self.unload_agent)
        self.unload_timer.start()


# Example usage
handler = ModelHandler()



import json
from pathlib import Path
import pathlib

DEFAULT_LAUNCH_SHORTCUT = "<cmd>+<shift>+<space>"
DEFAULT_SETTINGS = {
    "shortcut": DEFAULT_LAUNCH_SHORTCUT,
    "color": "#1E1E1E",
    "transparency": 95,
    "text_model": "hugging-quants/Llama-3.2-1B-Instruct-Q4_K_M-GGUF",
    "multimodal_model": "vikhyatk/moondream2",
    "hey_llama_chat": False,
    "hey_llama_mic": False,
    "generation": {"context_len": 2048, "top_k": 40, "top_p": 0.95, "temperature": 0.2},
    "rag": {
        "embed_model_name": "BAAI/bge-base-en-v1.5",
        "chunk_size": 256,
        "chunk_overlap": 128,
        "max_retrieval_top_k": 3,
        "similarity_threshold": 0.6,
    },
}
DEFAULT_MODELS = [
    {
        "model_name": "Llama-3.2-1B-Instruct-Q4_K_M-GGUF",
        "model_id": "hugging-quants/Llama-3.2-1B-Instruct-Q4_K_M-GGUF",
        "model_type": "text",
        "model_path": None,
        "repo_id": "hugging-quants/Llama-3.2-1B-Instruct-Q4_K_M-GGUF",
        "filename": "*q4_k_m.gguf",
    },
    {
        "model_name": "Llama-3.2-1B-Instruct-Q8_0-GGUF",
        "model_id": "hugging-quants/Llama-3.2-1B-Instruct-Q8_0-GGUF",
        "model_type": "text",
        "model_path": None,
        "repo_id": "hugging-quants/Llama-3.2-1B-Instruct-Q8_0-GGUF",
        "filename": "*q8_0.gguf",
    },
    {
        "model_name": "Llama-3.2-3B-Instruct-Q4_K_M-GGUF",
        "model_id": "hugging-quants/Llama-3.2-3B-Instruct-Q4_K_M-GGUF",
        "model_type": "text",
        "model_path": None,
        "repo_id": "hugging-quants/Llama-3.2-3B-Instruct-Q4_K_M-GGUF",
        "filename": "*q4_k_m.gguf",
    },
    {
        "model_name": "Llama-3.2-3B-Instruct-Q8_0-GGUF",
        "model_id": "hugging-quants/Llama-3.2-3B-Instruct-Q8_0-GGUF",
        "model_type": "text",
        "model_path": None,
        "repo_id": "hugging-quants/Llama-3.2-3B-Instruct-Q8_0-GGUF",
        "filename": "*q8_0.gguf",
    },
    {
        "model_name": "Qwen2.5-0.5B-Instruct-GGUF",
        "model_id": "Qwen/Qwen2.5-0.5B-Instruct-GGUF-q4_k_m",
        "model_type": "text",
        "model_path": None,
        "repo_id": "Qwen/Qwen2.5-0.5B-Instruct-GGUF",
        "filename": "*q4_k_m.gguf",
    },
    {
        "model_name": "Qwen2.5-1.5B-Instruct-GGUF",
        "model_id": "Qwen/Qwen2.5-1.5B-Instruct-GGUF-q4_k_m",
        "model_type": "text",
        "model_path": None,
        "repo_id": "Qwen/Qwen2.5-1.5B-Instruct-GGUF",
        "filename": "*q4_k_m.gguf",
    },
    {
        "model_name": "gemma-2-2b-it-GGUF-Q4_K_M",
        "model_id": "lmstudio-community/gemma-2-2b-it-GGUF-Q4_K_M",
        "model_type": "text",
        "model_path": None,
        "repo_id": "lmstudio-community/gemma-2-2b-it-GGUF",
        "filename": "gemma-2-2b-it-Q4_K_M.gguf",
    },
    {
        "model_name": "Moondream2",
        "model_id": "vikhyatk/moondream2",
        "model_type": "image",
        "model_path": None,
        "repo_id": "vikhyatk/moondream2",
        "filename": "*text-model*",
    },
    {
        "model_name": "Llava-1.5",
        "model_id": "mys/ggml_llava-v1.5-7b/q4_k",
        "model_type": "image",
        "model_path": None,
        "repo_id": "mys/ggml_llava-v1.5-7b",
        "filename": "*q4_k.gguf",
    },
    {
        "model_name": "Llava-1.5",
        "model_id": "mys/ggml_llava-v1.5-7b/f16",
        "model_type": "image",
        "model_path": None,
        "repo_id": "mys/ggml_llava-v1.5-7b",
        "filename": "*f16.gguf",
    },
    {
        "model_name": "MiniCPM-V-2_6-gguf",
        "model_id": "openbmb/MiniCPM-V-2_6-gguf-Q4_K_M",
        "model_type": "image",
        "model_path": None,
        "repo_id": "openbmb/MiniCPM-V-2_6-gguf",
        "filename": "*Q4_K_M.gguf",
    },
    {
        "model_name": "MiniCPM-V-2_6-gguf",
        "model_id": "openbmb/MiniCPM-V-2_6-gguf-Q8_0",
        "model_type": "image",
        "model_path": None,
        "repo_id": "openbmb/MiniCPM-V-2_6-gguf",
        "filename": "*Q8_0.gguf",
    },
]

# generation setting
context_len = 2048
top_k = 40
top_p = 0.95
temperature = 0.2

home_dir = Path.home()
llama_assistant_dir = home_dir / "llama_assistant"
pathlib.Path.mkdir(llama_assistant_dir, parents=True, exist_ok=True)
custom_models_file = llama_assistant_dir / "custom_models.json"
settings_file = llama_assistant_dir / "settings.json"
document_icon = "llama_assistant/resources/document_icon.png"

# RAG setting
embed_model_name = "BAAI/bge-base-en-v1.5"
chunk_size = 256
chunk_overlap = 128
max_retrieval_top_k = 3
similarity_threshold = 0.6

if custom_models_file.exists():
    with open(custom_models_file, "r") as f:
        try:
            config_data = json.load(f)
        except json.JSONDecodeError:
            config_data = {"custom_models": []}
    custom_models = config_data.get("custom_models", [])
else:
    custom_models = []

models = DEFAULT_MODELS + custom_models

# Save the initial configuration if it doesn't exist
if not custom_models_file.exists():
    config_dir = custom_models_file.parent
    if not config_dir.exists():
        config_dir.mkdir(parents=True)
    with open(custom_models_file, "w") as f:
        json.dump({"custom_models": custom_models}, f, indent=2)


# Save the custom models to the file
def save_custom_models():
    global models
    with open(custom_models_file, "w") as f:
        json.dump({"custom_models": custom_models}, f, indent=2)
    models = DEFAULT_MODELS + custom_models



import sys
import multiprocessing
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import Qt
from llama_assistant.llama_assistant_app import LlamaAssistant


def main():
    # Enable high DPI scaling
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

    app = QApplication(sys.argv)
    ex = LlamaAssistant()
    ex.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



import base64
import os
import sys
from importlib import resources

from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import Qt


def image_to_base64_data_uri(file_path):
    with open(file_path, "rb") as img_file:
        base64_data = base64.b64encode(img_file.read()).decode("utf-8")
        return f"data:image/png;base64,{base64_data}"


def get_resource_path(relative_path):
    """Get the path to a resource, works for dev and for PyInstaller"""
    if getattr(sys, "frozen", False):
        # Running in a bundle (PyInstaller)
        base_path = sys._MEIPASS
    else:
        # Running in normal Python environment
        base_path = os.path.dirname(os.path.abspath(__file__))

    path = os.path.join(base_path, relative_path)
    if not os.path.exists(path):
        # Fallback to reading from the package
        file_name = os.path.basename(path)
        package_name = os.path.dirname(relative_path).replace("/", ".")
        with resources.path(package_name, file_name) as p:
            path = str(p)

    return path


def load_image(relative_path, size=None):
    """Load an image from a relative path and optionally resize it"""
    full_path = get_resource_path(relative_path)
    if os.path.exists(full_path):
        pixmap = QPixmap(full_path)
        if size:
            pixmap = pixmap.scaled(size[0], size[1], Qt.AspectRatioMode.KeepAspectRatio)
        return pixmap
    else:
        print(f"Image not found at: {full_path}")
        return QPixmap()



import time
import os
import re
import traceback

from PyQt5.QtCore import QThread, pyqtSignal
import pyaudio
import wave

from whispercpp import Whisper
from llama_assistant.config import llama_assistant_dir


class SpeechRecognitionThread(QThread):
    finished = pyqtSignal(str)
    error = pyqtSignal(str)
    WHISPER_THREADS = 1
    MAX_RECORDING_TIME = 60  # Maximum recording time in seconds

    def __init__(self):
        super().__init__()
        self.stop_listening = False
        self.recording = False

        # Initialize Whisper model
        self.whisper = Whisper("tiny")

        # Create temporary folder for audio files
        self.tmp_audio_folder = llama_assistant_dir / "tmp_audio"
        self.tmp_audio_folder.mkdir(parents=True, exist_ok=True)

        # Audio recording parameters
        self.FORMAT = pyaudio.paInt16
        self.CHANNELS = 1
        self.RATE = 16000
        self.CHUNK = 1024

    def wait_for_recording(self):
        while not self.recording:
            time.sleep(0.1)

    def run(self):
        self.stop_listening = False
        audio = pyaudio.PyAudio()
        frames = []
        start_time = time.time()

        try:
            stream = audio.open(
                format=self.FORMAT,
                channels=self.CHANNELS,
                rate=self.RATE,
                input=True,
                frames_per_buffer=self.CHUNK,
            )

            print("Always-on microphone activated. Listening...")
            self.recording = True

            while not self.stop_listening and (time.time() - start_time) < self.MAX_RECORDING_TIME:
                data = stream.read(self.CHUNK)
                frames.append(data)

            self.recording = False
            print("Stopped recording. Processing audio...")

            # Save audio data to temporary file
            tmp_filepath = self.tmp_audio_folder / f"temp_audio_{time.time()}.wav"
            wf = wave.open(str(tmp_filepath), "wb")
            wf.setnchannels(self.CHANNELS)
            wf.setsampwidth(audio.get_sample_size(self.FORMAT))
            wf.setframerate(self.RATE)
            wf.writeframes(b"".join(frames))
            wf.close()

            # Transcribe audio
            res = self.whisper.transcribe(str(tmp_filepath))
            transcription = self.whisper.extract_text(res)

            if isinstance(transcription, list):
                # Remove all "[BLANK_AUDIO]" from the transcription
                transcription = " ".join(transcription)
                transcription = re.sub(r"\[BLANK_AUDIO\]", "", transcription)

            if transcription.strip():  # Only emit if there's non-empty transcription
                self.finished.emit(transcription)

            os.remove(tmp_filepath)

        except Exception as e:
            traceback.print_exc()
            self.error.emit(f"An error occurred: {str(e)}")
        finally:
            stream.stop_stream()
            stream.close()
            audio.terminate()

    def stop(self):
        self.stop_listening = True
        print("Stopping always-on microphone...")


# Updated demo code
if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QLabel, QPushButton
    import sys

    class DemoWidget(QWidget):
        def __init__(self):
            super().__init__()
            self.init_ui()
            self.thread = SpeechRecognitionThread()
            self.thread.finished.connect(self.on_finished)
            self.thread.error.connect(self.on_error)

        def init_ui(self):
            layout = QVBoxLayout()
            self.status_label = QLabel("Click 'Start' to begin always-on speech recognition")
            self.transcription_label = QLabel("Transcription will appear here")
            self.start_button = QPushButton("Start")
            self.stop_button = QPushButton("Stop")
            self.stop_button.setEnabled(False)

            layout.addWidget(self.status_label)
            layout.addWidget(self.transcription_label)
            layout.addWidget(self.start_button)
            layout.addWidget(self.stop_button)

            self.setLayout(layout)
            self.setWindowTitle("Always-On Speech Recognition Demo")

            self.start_button.clicked.connect(self.start_recognition)
            self.stop_button.clicked.connect(self.stop_recognition)

        def start_recognition(self):
            self.status_label.setText("Always-on microphone activated. Listening...")
            self.start_button.setEnabled(False)
            self.stop_button.setEnabled(True)
            self.thread.start()

        def stop_recognition(self):
            self.thread.stop()
            self.status_label.setText("Processing recorded audio...")
            self.start_button.setEnabled(False)
            self.stop_button.setEnabled(False)

        def on_finished(self, text):
            self.transcription_label.setText(f"Transcription: {text}")
            self.status_label.setText("Speech recognition completed")
            self.start_button.setEnabled(True)
            self.stop_button.setEnabled(False)

        def on_error(self, error_message):
            self.status_label.setText(f"Error: {error_message}")
            self.start_button.setEnabled(True)
            self.stop_button.setEnabled(False)

    app = QApplication(sys.argv)
    demo = DemoWidget()
    demo.show()
    sys.exit(app.exec_())



from typing import List, Set, Optional, Dict, TYPE_CHECKING

from llama_cpp import Llama
from llama_index.core import VectorStoreIndex
from llama_index.embeddings.huggingface import HuggingFaceEmbedding
from llama_index.core.schema import NodeWithScore
from llama_index.core import VectorStoreIndex, SimpleDirectoryReader, Settings
from llama_index.core.workflow import Context
from llama_index.core.postprocessor import SimilarityPostprocessor

from llama_index.core.workflow import Event, StartEvent, StopEvent, Workflow, step


def convert_message_list_to_str(messages):
    chat_history_str = ""
    for message in messages:
        if type(message["content"]) is str:
            chat_history_str += message["role"] + ": " + message["content"] + "\n"
        else:
            chat_history_str += message["role"] + ": " + message["content"]["text"] + "\n"

    return chat_history_str


class SetupEvent(Event):
    pass


class CondenseQueryEvent(Event):
    condensed_query_str: str


class RetrievalEvent(Event):
    nodes: List[NodeWithScore]


class ChatHistory:
    def __init__(self, max_history_size: int):
        self.max_history_size = max_history_size
        self.total_size = 0
        self.chat_history = []

    def add_message(self, message: dict):
        if "content" in message and type(message["content"]) is list:
            # multimodal model's message format
            new_msg_size = len(message["content"][0]["text"].split())
        else:
            # text-only model's message format
            new_msg_size = len(message["content"].split())

        self.total_size += new_msg_size
        self.chat_history.append(message)
        while self.total_size > self.max_history_size:
            oldest_msg = self.chat_history.pop(0)
            if "content" in oldest_msg and type(oldest_msg["content"]) is list:
                # multimodal model's message format
                len_oldest_msg = len(oldest_msg["content"][0]["text"].split())
            else:
                # text-only model's message format
                len_oldest_msg = len(oldest_msg["content"].split())
            self.total_size -= len_oldest_msg

    def get_chat_history(self):
        return self.chat_history

    def clear(self):
        self.chat_history = []
        self.total_size = 0

    def __len__(self):
        return len(self.chat_history)


class RAGAgent(Workflow):
    SUMMARY_TEMPLATE = (
        "Given the converstation:\n"
        "'''{chat_history_str}'''\n\n"
        "Rewrite to a standalone question:\n"
    )

    CONTEXT_PROMPT_TEMPLATE = (
        "Information that might help:\n"
        "-----\n"
        "{node_context}\n"
        "-----\n"
        "Please write a response to the following question, using the above information if relevant:\n"
        "{query_str}\n"
    )
    SYSTEM_PROMPT = {"role": "system", "content": "Generate short and simple response."}

    def __init__(
        self,
        generation_setting: Dict,
        rag_setting: Dict,
        llm: Llama,
        timeout: int = 60,
        verbose: bool = False,
    ):
        super().__init__(timeout=timeout, verbose=verbose)
        self.generation_setting = generation_setting
        self.context_len = generation_setting["context_len"]
        # we want the retrieved context = our set value but not more than the context_len
        self.retrieval_top_k = (self.context_len - rag_setting["chunk_size"]) // rag_setting[
            "chunk_overlap"
        ]
        self.retrieval_top_k = min(max(1, self.retrieval_top_k), rag_setting["max_retrieval_top_k"])
        self.search_index = None
        self.retriever = None
        # 1 token ~ 3/4 words, because the context length accounts for both input and output tokens
        # we need to reserve some space for the output tokens (half-space for output tokens)
        self.chat_history = ChatHistory(max_history_size=self.context_len * 0.75 * 1 / 2)
        self.lookup_files = set()

        self.embed_model = HuggingFaceEmbedding(model_name=rag_setting["embed_model_name"])
        Settings.embed_model = self.embed_model
        Settings.chunk_size = rag_setting["chunk_size"]
        Settings.chunk_overlap = rag_setting["chunk_overlap"]
        self.node_processor = SimilarityPostprocessor(
            similarity_cutoff=rag_setting["similarity_threshold"]
        )
        self.llm = llm

    def update_index(self, files: Optional[Set[str]] = set()):
        if not files:
            print("No lookup files provided, clearing index...")
            self.retriever = None
            self.search_index = None
            return

        print("Indexing documents...")
        documents = SimpleDirectoryReader(input_files=files, recursive=True).load_data(
            show_progress=True, num_workers=1
        )

        if self.search_index is None:
            self.search_index = VectorStoreIndex.from_documents(
                documents, embed_model=self.embed_model
            )
        else:
            for doc in documents:
                self.search_index.insert(doc)  # Add the new document to the index

        self.retriever = self.search_index.as_retriever(similarity_top_k=self.retrieval_top_k)

    def update_rag_setting(self, rag_setting: Dict):
        if self.node_processor.similarity_cutoff != rag_setting["similarity_threshold"]:
            self.node_processor = SimilarityPostprocessor(
                similarity_cutoff=rag_setting["similarity_threshold"]
            )

        new_top_k = (self.context_len - rag_setting["chunk_size"]) // rag_setting["chunk_overlap"]
        new_top_k = min(max(1, new_top_k), rag_setting["max_retrieval_top_k"])
        if self.retrieval_top_k != new_top_k:
            self.retrieval_top_k = new_top_k
            if self.retriever:
                self.retriever = self.search_index.as_retriever(
                    similarity_top_k=self.retrieval_top_k
                )

        if (
            self.embed_model.model_name != rag_setting["embed_model_name"]
            or Settings.chunk_size != rag_setting["chunk_size"]
            or Settings.chunk_overlap != rag_setting["chunk_overlap"]
        ):
            self.embed_model = HuggingFaceEmbedding(model_name=rag_setting["embed_model_name"])
            Settings.embed_model = self.embed_model
            Settings.chunk_size = rag_setting["chunk_size"]
            Settings.chunk_overlap = rag_setting["chunk_overlap"]

            # reindex since those are the settings that affect the index
            if self.lookup_files:
                print("Re-indexing documents since the rag settings have changed...")
                documents = SimpleDirectoryReader(
                    input_files=self.lookup_files, recursive=True
                ).load_data(show_progress=True, num_workers=1)
                self.search_index = VectorStoreIndex.from_documents(
                    documents, embed_model=self.embed_model
                )
                self.retriever = self.search_index.as_retriever(
                    similarity_top_k=self.retrieval_top_k
                )

    def update_generation_setting(self, generation_setting):
        self.generation_setting = generation_setting
        self.context_len = generation_setting["context_len"]

    @step
    async def setup(self, ctx: Context, ev: StartEvent) -> SetupEvent:
        # set frequently used variables to context
        query_str = ev.query_str
        image = ev.image
        lookup_files = ev.lookup_files if ev.lookup_files else set()
        streaming = ev.streaming
        await ctx.set("query_str", query_str)
        await ctx.set("image", image)
        await ctx.set("streaming", streaming)

        # update index if needed
        if lookup_files != self.lookup_files:
            print("Different lookup files, updating index...")
            self.update_index(lookup_files)

        self.lookup_files = lookup_files.copy()

        return SetupEvent()

    @step
    async def condense_history_to_query(self, ctx: Context, ev: SetupEvent) -> CondenseQueryEvent:
        """
        Condense the chat history and the query into a single query. Only used for retrieval.
        """
        query_str = await ctx.get("query_str")

        formated_query = ""

        if len(self.chat_history) > 0 or self.retriever is not None:
            self.chat_history.add_message({"role": "user", "content": query_str})
            chat_history_str = convert_message_list_to_str(self.chat_history.get_chat_history())
            # use llm to summarize the chat history into a single query,
            # which is used to retrieve context from the documents
            formated_query = self.SUMMARY_TEMPLATE.format(chat_history_str=chat_history_str)

            history_summary = self.llm.create_chat_completion(
                messages=[{"role": "user", "content": formated_query}],
                stream=False,
                top_k=self.generation_setting["top_k"],
                top_p=self.generation_setting["top_p"],
                temperature=self.generation_setting["temperature"],
            )["choices"][0]["message"]["content"]

            condensed_query = "Context:\n" + history_summary + "\nQuestion: " + query_str
            # remove the last user message from the chat history
            # later we will add the query with retrieved context (RAG)
            self.chat_history.chat_history.pop()
        else:
            # if there is no history or no need for retrieval, return the query as is
            condensed_query = query_str

        return CondenseQueryEvent(condensed_query_str=condensed_query)

    @step
    async def retrieve(self, ctx: Context, ev: CondenseQueryEvent) -> RetrievalEvent:
        if not self.retriever:
            return RetrievalEvent(nodes=[])

        # retrieve from dropped documents
        condensed_query_str = ev.condensed_query_str
        nodes = await self.retriever.aretrieve(condensed_query_str)
        nodes = self.node_processor.postprocess_nodes(nodes)
        return RetrievalEvent(nodes=nodes)

    def _prepare_query_with_context(
        self,
        query_str: str,
        nodes: List[NodeWithScore],
    ) -> str:
        node_context = ""

        if len(nodes) == 0:
            return query_str

        for idx, node in enumerate(nodes):
            node_text = node.get_content(metadata_mode="llm")
            node_context += f"\n{node_text}\n\n"

        formatted_query = self.CONTEXT_PROMPT_TEMPLATE.format(
            node_context=node_context, query_str=query_str
        )

        return formatted_query

    @step
    async def llm_response(self, ctx: Context, retrieval_ev: RetrievalEvent) -> StopEvent:
        nodes = retrieval_ev.nodes
        query_str = await ctx.get("query_str")
        image = await ctx.get("image")
        query_with_ctx = self._prepare_query_with_context(query_str, nodes)
        streaming = await ctx.get("streaming", False)

        if image:
            formated_message = {
                "role": "user",
                "content": [
                    {"type": "text", "text": query_with_ctx},
                    {"type": "image_url", "image_url": {"url": image}},
                ],
            }
        else:
            formated_message = {"role": "user", "content": query_with_ctx}

        self.chat_history.add_message(formated_message)

        response = self.llm.create_chat_completion(
            messages=[self.SYSTEM_PROMPT] + self.chat_history.get_chat_history(),
            stream=streaming,
            top_k=self.generation_setting["top_k"],
            top_p=self.generation_setting["top_p"],
            temperature=self.generation_setting["temperature"],
        )
        # remove the query with context from the chat history,
        self.chat_history.chat_history.pop()
        # add the short query (without context) instead -> not to clutter the chat history
        self.chat_history.add_message({"role": "user", "content": query_str})

        return StopEvent(result=response)



import json
import copy
import time
import traceback
import mistune

from PyQt5.QtWidgets import (
    QApplication,
    QMainWindow,
    QPushButton,
    QLabel,
    QWidget,
    QVBoxLayout,
    QMessageBox,
    QSystemTrayIcon,
)
from PyQt5.QtCore import (
    Qt,
    QPoint,
    QTimer,
)
from PyQt5.QtGui import (
    QPixmap,
    QPainter,
    QDragEnterEvent,
    QDropEvent,
    QBitmap,
    QTextCursor,
    QFont,
)

from llama_assistant import config
from llama_assistant.wake_word_detector import WakeWordDetector
from llama_assistant.global_hotkey import GlobalHotkey
from llama_assistant.setting_dialog import SettingsDialog
from llama_assistant.speech_recognition_thread import SpeechRecognitionThread
from llama_assistant.utils import image_to_base64_data_uri
from llama_assistant.processing_thread import ProcessingThread
from llama_assistant.ui_manager import UIManager
from llama_assistant.tray_manager import TrayManager


class LlamaAssistant(QMainWindow):
    def __init__(self):
        super().__init__()
        self.wake_word_detector = None
        self.load_settings()
        self.ui_manager = UIManager(self)
        self.tray_manager = TrayManager(self)
        self.setup_global_shortcut()
        self.last_response = ""
        self.dropped_image = None
        self.dropped_files = set()
        self.file_containers = dict()
        self.speech_thread = None
        self.is_listening = False
        self.image_label = None
        self.current_text_model = self.settings.get("text_model")
        self.current_multimodal_model = self.settings.get("multimodal_model")
        self.processing_thread = None
        self.markdown_creator = mistune.create_markdown()

    def tray_icon_activated(self, reason):
        if reason == QSystemTrayIcon.ActivationReason.Trigger:
            self.show()
            self.activateWindow()
            self.raise_()

    def init_wake_word_detector(self):
        if self.wake_word_detector is not None:
            self.deinit_wake_word_detector()
        self.wake_word_detector = WakeWordDetector()
        self.wake_word_detector.wakeword_detected.connect(self.on_wake_word_detected)
        self.wake_word_detector.start()

    def deinit_wake_word_detector(self):
        if self.wake_word_detector.running:
            self.wake_word_detector.stop()
        self.wake_word_detector = None

    def load_settings(self):
        if config.settings_file.exists():
            with open(config.settings_file, "r") as f:
                self.settings = json.load(f)
            self.settings["text_model"] = self.settings.get(
                "text_model", config.DEFAULT_SETTINGS["text_model"]
            )
            self.settings["multimodal_model"] = self.settings.get(
                "multimodal_model", config.DEFAULT_SETTINGS["multimodal_model"]
            )
            # Update any missing keys from default settings
            for key, default_value in config.DEFAULT_SETTINGS.items():
                if key not in self.settings:
                    self.settings[key] = default_value
            self.save_settings()
        else:
            self.settings = copy.deepcopy(config.DEFAULT_SETTINGS)
            self.save_settings()
        if self.settings.get("hey_llama_chat", False) and self.wake_word_detector is None:
            self.init_wake_word_detector()
        if not self.settings.get("hey_llama_chat", False) and self.wake_word_detector is not None:
            self.deinit_wake_word_detector()
        self.current_text_model = self.settings.get("text_model")
        self.current_multimodal_model = self.settings.get("multimodal_model")
        self.generation_setting = self.settings.get("generation")
        self.rag_setting = self.settings.get("rag")

    def setup_global_shortcut(self):
        try:
            if hasattr(self, "global_hotkey"):
                self.global_hotkey.stop()
                time.sleep(0.1)  # Give a short delay to ensure the previous listener has stopped
            try:
                self.global_hotkey = GlobalHotkey(self.settings["shortcut"])
                self.global_hotkey.activated.connect(self.toggle_visibility)
            except Exception as e:
                print(f"Error setting up global shortcut: {e}")
                # Fallback to default shortcut if there's an error
                self.global_hotkey = GlobalHotkey(config.DEFAULT_LAUNCH_SHORTCUT)
                self.global_hotkey.activated.connect(self.toggle_visibility)
        except Exception as e:
            print(f"Error setting up global shortcut: {e}")
            traceback.print_exc()

    def open_settings(self):
        dialog = SettingsDialog(self)
        if dialog.exec():
            new_settings = dialog.get_settings()
            old_shortcut = self.settings["shortcut"]
            self.settings.update(new_settings)
            self.save_settings()
            self.load_settings()
            self.ui_manager.update_styles()

            if old_shortcut != self.settings["shortcut"]:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Icon.Information)
                msg.setText("Global shortcut has been updated")
                msg.setInformativeText(
                    "The changes will take effect after you restart the application."
                )
                msg.setWindowTitle("Restart Required")
                msg.setStandardButtons(
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
                )
                msg.button(QMessageBox.StandardButton.Yes).setText("Restart Now")
                msg.button(QMessageBox.StandardButton.No).setText("Restart Later")
                msg.setDefaultButton(QMessageBox.StandardButton.Yes)

                result = msg.exec()

                if result == QMessageBox.StandardButton.Yes:
                    self.restart_application()
                else:
                    print("User chose to restart later.")

    def restart_application(self):
        QApplication.quit()
        # The application will restart automatically because it is being run from a script

    def save_settings(self):
        with open(config.settings_file, "w") as f:
            json.dump(self.settings, f)

    def toggle_visibility(self):
        if self.isVisible():
            self.hide()
        else:
            self.show()
            self.activateWindow()
            self.raise_()
            self.ui_manager.input_field.setFocus()

    def on_submit(self):
        message = self.ui_manager.input_field.toPlainText()
        if message == "":
            return
        self.ui_manager.input_field.clear()

        if message == "cls" or message == "clear":
            self.clear_chat()
            self.remove_image_thumbnail()
            self.dropped_image = None

            for file_path in self.dropped_files:
                self.remove_file_thumbnail(self.file_containers[file_path], file_path)

            return

        self.last_response = ""

        if self.dropped_image:
            self.process_image_with_prompt(self.dropped_image, self.dropped_files, message)
            self.dropped_image = None
            self.remove_image_thumbnail()
        else:
            QTimer.singleShot(100, lambda: self.process_text(message, self.dropped_files, "chat"))

    def on_task_button_clicked(self):
        button = self.sender()
        task = button.text()
        message = self.ui_manager.input_field.toPlainText()
        if message == "":
            return
        self.process_text(message, self.dropped_files, task)

    def process_text(self, message, file_paths, task="chat"):
        if task != "chat":
            self.clear_chat()
        self.show_chat_box()
        if task == "chat":
            prompt = message
        elif task == "Summarize":
            prompt = f"Summarize the following text: {message}"
        elif task == "Rephrase":
            prompt = f"Rephrase the following text {message}"
        elif task == "Fix Grammar":
            prompt = f"Fix the grammar in the following text:\n {message}"
        elif task == "Brainstorm":
            prompt = f"Brainstorm ideas related to: {message}"
        elif task == "Write Email":
            prompt = f"Write an email about: {message}"

        self.ui_manager.chat_box.append(f'<span style="color: #aaa;"><b>You:</b></span> {message}')
        self.ui_manager.chat_box.append(f'<span style="color: #aaa;"><b>AI ({task}):</b></span> ')

        self.start_cursor_pos = self.ui_manager.chat_box.textCursor().position()

        self.processing_thread = ProcessingThread(
            self.current_text_model,
            self.generation_setting,
            self.rag_setting,
            prompt,
            lookup_files=file_paths,
        )

        self.processing_thread.preloader_signal.connect(self.indicate_loading)
        self.processing_thread.update_signal.connect(self.update_chat_box)
        self.processing_thread.finished_signal.connect(self.on_processing_finished)
        self.processing_thread.start()

    def process_image_with_prompt(self, image_path, file_paths, prompt):
        self.show_chat_box()
        self.ui_manager.chat_box.append(
            f'<span style="color: #aaa;"><b>You:</b></span> [Uploaded an image: {image_path}]'
        )
        self.ui_manager.chat_box.append(f'<span style="color: #aaa;"><b>You:</b></span> {prompt}')
        self.ui_manager.chat_box.append('<span style="color: #aaa;"><b>AI:</b></span> ')

        self.ui_manager.chat_box.moveCursor(QTextCursor.End)
        self.start_cursor_pos = self.ui_manager.chat_box.textCursor().position()

        image = image_to_base64_data_uri(image_path)
        self.processing_thread = ProcessingThread(
            self.current_multimodal_model,
            self.generation_setting,
            self.rag_setting,
            prompt,
            image=image,
            lookup_files=file_paths,
        )
        self.processing_thread.preloader_signal.connect(self.indicate_loading)
        self.processing_thread.update_signal.connect(self.update_chat_box)
        self.processing_thread.finished_signal.connect(self.on_processing_finished)
        self.processing_thread.start()

    def indicate_loading(self, message):
        while self.processing_thread.is_preloading():
            cursor = self.ui_manager.chat_box.textCursor()
            cursor.setPosition(self.start_cursor_pos)
            # Select all text from the start_pos to the end
            cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor)
            # Remove the selected text
            cursor.removeSelectedText()
            # create animation where the characters are displayed one by one
            for c in message:
                if c == " ":
                    cursor.insertText(" ")
                else:
                    cursor.insertHtml(f'<span style="color: #aaa;">{c}</span>')
                QApplication.processEvents()  # Process events to update the UI
                time.sleep(0.05)
            time.sleep(0.5)

    def update_chat_box(self, text):
        self.last_response += text
        markdown_response = self.markdown_creator(self.last_response)
        # Since cannot change the font size of the h1, h2 tag, we will replace it with h3
        markdown_response = markdown_response.replace("<h1>", "<h3>").replace("</h1>", "</h3>")
        markdown_response = markdown_response.replace("<h2>", "<h3>").replace("</h2>", "</h3>")
        markdown_response += "<div></div>"
        cursor = self.ui_manager.chat_box.textCursor()
        cursor.setPosition(
            self.start_cursor_pos
        )  # regenerate the updated text from the start position
        # Select all text from the start_pos to the end
        cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor)
        # Remove the selected text
        cursor.removeSelectedText()
        cursor.insertHtml(markdown_response)
        self.ui_manager.chat_box.verticalScrollBar().setValue(
            self.ui_manager.chat_box.verticalScrollBar().maximum()
        )

    def on_processing_finished(self):
        self.ui_manager.chat_box.textCursor().movePosition(QTextCursor.End)

    def show_chat_box(self):
        if self.ui_manager.scroll_area.isHidden():
            self.ui_manager.scroll_area.show()
            self.ui_manager.copy_button.show()
            self.ui_manager.clear_button.show()
            self.setFixedHeight(700)  # Increase this value if needed

        self.ui_manager.chat_box.verticalScrollBar().setValue(
            self.ui_manager.chat_box.verticalScrollBar().maximum()
        )

    def copy_result(self):
        self.hide()
        if self.last_response:
            clipboard = QApplication.clipboard()
            clipboard.setText(self.last_response)

    def clear_chat(self):
        self.ui_manager.chat_box.clear()
        self.last_response = ""
        self.ui_manager.scroll_area.hide()
        self.ui_manager.input_field.clear()
        self.ui_manager.input_field.setFocus()
        self.ui_manager.copy_button.hide()
        self.ui_manager.clear_button.hide()
        self.processing_thread.clear_chat_history()
        self.setFixedHeight(400)  # Reset to default height

    def dragEnterEvent(self, event: QDragEnterEvent):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event: QDropEvent):
        files = [u.toLocalFile() for u in event.mimeData().urls()]
        for file_path in files:
            print(f"File dropped: {file_path}")
            if file_path.lower().endswith((".png", ".jpg", ".jpeg", ".gif", ".bmp")):
                self.dropped_image = file_path
                self.ui_manager.input_field.setPlaceholderText("Enter a prompt for the image...")
                self.show_image_thumbnail(file_path)
            elif file_path.lower().endswith((".pdf", "doc", ".docx", ".txt")):
                if file_path not in self.dropped_files:
                    self.dropped_files.add(file_path)
                    self.ui_manager.input_field.setPlaceholderText(
                        "Enter a prompt for the document..."
                    )
                    self.show_file_thumbnail(file_path)
                else:
                    print(f"File {file_path} already added")

    def remove_file_thumbnail(self, file_label, file_path):
        file_label.setParent(None)
        self.setFixedHeight(self.height() - 110)  # Decrease height after removing file
        # Remove the file from the list
        self.dropped_files.remove(file_path)
        del self.file_containers[file_path]

    def show_image_thumbnail(self, image_path):
        if self.image_label is None:
            self.image_label = QLabel(self)
            self.image_label.setFixedSize(80, 80)
            self.image_label.setStyleSheet(
                """
                background-color: transparent;
                """
            )

            remove_button = QPushButton("×", self.image_label)
            remove_button.setStyleSheet(
                """
                QPushButton {
                    background-color: rgba(50, 50, 50, 200);
                    color: white;
                    border: none;
                    border-radius: 8px;
                    font-size: 12px;
                    padding: 2px;
                    width: 16px;
                    height: 16px;
                }
                QPushButton:hover {
                    background-color: rgba(50, 50, 50, 230);
                }
                """
            )
            remove_button.move(60, 0)
            remove_button.clicked.connect(self.remove_image_thumbnail)

        pixmap = QPixmap(image_path)
        scaled_pixmap = pixmap.scaled(
            80,
            80,
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )

        # Create a rounded mask
        mask = QBitmap(scaled_pixmap.size())
        mask.fill(Qt.GlobalColor.color0)
        painter = QPainter(mask)
        painter.setBrush(Qt.GlobalColor.color1)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.drawRoundedRect(mask.rect(), 20, 20)
        painter.end()

        # Apply the mask to the pixmap
        rounded_pixmap = scaled_pixmap.copy()
        rounded_pixmap.setMask(mask)

        self.image_label.setPixmap(rounded_pixmap)

        # Clear previous image if any
        for i in reversed(range(self.ui_manager.image_layout.count())):
            self.ui_manager.image_layout.itemAt(i).widget().setParent(None)

        # Add new image to layout
        self.ui_manager.image_layout.addWidget(self.image_label)
        self.setFixedHeight(self.height() + 110)  # Increase height to accommodate larger image

    def show_file_thumbnail(self, file_path):
        # Create a container widget
        container = QWidget(self)
        container.setFixedSize(80, 100)  # Adjust height to accommodate both pixmap and text

        # Create a layout for the container
        layout = QVBoxLayout(container)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        # Create a QLabel for the pixmap
        pixmap_label = QLabel(container)
        pixmap_label.setFixedSize(80, 80)
        pixmap_label.setStyleSheet("background-color: transparent;")

        # Create a QLabel for the text
        text_label = QLabel(file_path.split("/")[-1], container)
        # set text background color to white, text size to 5px
        #  and rounded corners, vertical alignment to top
        text_label.setStyleSheet(
            """
            background-color: black;
            color: white;
            border-radius: 5px;
            font-size: 8px;
            padding: 2px;
            """
        )
        text_label.setWordWrap(True)
        text_label.setAlignment(Qt.AlignmentFlag.AlignTop)

        # Add the labels to the layout
        layout.addWidget(pixmap_label)
        layout.addWidget(text_label)

        # Create the remove button
        remove_button = QPushButton("x", pixmap_label)
        remove_button.setStyleSheet(
            """
            QPushButton {
                background-color: rgba(50, 50, 50, 200);
                color: white;
                border: none;
                border-radius: 8px;
                font-size: 12px;
                padding: 2px;
                width: 16px;
                height: 16px;
            }
            QPushButton:hover {
                background-color: rgba(50, 50, 50, 230);
            }
            """
        )
        remove_button.move(60, 0)
        remove_button.clicked.connect(lambda: self.remove_file_thumbnail(container, file_path))

        # Load and set the pixmap
        import os

        print("Icon path:", str(config.document_icon), os.path.exists(str(config.document_icon)))
        pixmap = QPixmap(str(config.document_icon))
        scaled_pixmap = pixmap.scaled(
            80,
            80,
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )
        pixmap_label.setPixmap(scaled_pixmap)

        # Add the container to the layout
        self.ui_manager.file_layout.addWidget(container)

        self.setFixedHeight(self.height() + 110)  # Increase height to accommodate larger file

        self.file_containers[file_path] = container

    def remove_image_thumbnail(self):
        if self.image_label:
            self.image_label.setParent(None)
            self.image_label = None
            self.dropped_image = None
            self.ui_manager.input_field.setPlaceholderText("Ask me anything...")
            self.setFixedHeight(self.height() - 110)  # Decrease height after removing image

    def mousePressEvent(self, event):
        self.oldPos = event.globalPos()

    def mouseMoveEvent(self, event):
        delta = QPoint(event.globalPos() - self.oldPos)
        self.move(self.x() + delta.x(), self.y() + delta.y())
        self.oldPos = event.globalPos()

    def on_wake_word_detected(self, model_name):
        self.show()
        self.activateWindow()
        self.raise_()
        if self.settings.get("hey_llama_mic", False):
            self.start_voice_input()

    def toggle_voice_input(self):
        if not self.is_listening:
            self.start_voice_input()
        else:
            self.stop_voice_input()

    def start_voice_input(self):
        if self.speech_thread is None or not self.speech_thread.isRunning():
            self.is_listening = True
            self.ui_manager.mic_button.setStyleSheet(
                """
                QPushButton {
                    background-color: rgba(240, 150, 20, 0.5);
                    border: none;
                    border-radius: 20px;
                }
                QPushButton:hover {
                    background-color: rgba(240, 150, 20, 0.6);
                }
            """
            )
            self.speech_thread = SpeechRecognitionThread()
            self.speech_thread.finished.connect(self.on_speech_recognized)
            self.speech_thread.error.connect(self.on_speech_error)
            self.speech_thread.start()

            # Use QTimer to delay the application of the second style
            QTimer.singleShot(500, self.update_mic_button_style)

    def update_mic_button_style(self):
        self.ui_manager.mic_button.setStyleSheet(
            """
            QPushButton {
                background-color: rgba(255, 0, 0, 0.5);
                border: none;
                border-radius: 20px;
            }
            QPushButton:hover {
                background-color: rgba(255, 0, 0, 0.6);
            }
        """
        )

    def stop_voice_input(self):
        if self.speech_thread and self.speech_thread.isRunning():
            self.is_listening = False
            self.speech_thread.stop()
            self.ui_manager.mic_button.setStyleSheet(
                """
                QPushButton {
                    background-color: rgba(255, 255, 255, 0.5);
                    border: none;
                    border-radius: 20px;
                }
                QPushButton:hover {
                    background-color: rgba(255, 255, 255, 0.6);
                }
            """
            )

    def on_speech_recognized(self, text):
        current_text = self.ui_manager.input_field.toPlainText()
        if current_text:
            self.ui_manager.input_field.setPlainText(f"{current_text}\n{text}")
        else:
            self.ui_manager.input_field.setPlainText(text)
        self.stop_voice_input()

    def on_speech_error(self, error_message):
        print(f"Speech recognition error: {error_message}")
        self.stop_voice_input()

    def closeEvent(self, event):
        if self.wake_word_detector is not None:
            self.wake_word_detector.stop()
        super().closeEvent(event)
